/*
 * Copyright (C) 2010 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Mikkel Kamstrup Erlandsen <mikkel.kamstrup@canonical.com>
 *
 */
using Dee;
using DBus;
using Zeitgeist;
using Zeitgeist.Timestamp;
using Unity.Place;
using Config;
using Gee;
using GMenu;

namespace Unity.ApplicationsPlace {

  public class Daemon : GLib.Object
  {
    private Zeitgeist.Log log;
    private Zeitgeist.Index index;
    private Unity.Package.Searcher pkgsearcher;

    private Unity.Place.Controller control;
    private Unity.Place.EntryInfo applications;

    /* For each section we have a set filtering query we use to restrict
     * Xapian queries to that section */
    private Gee.List<string> section_queries;

    private Gee.List<string> image_extensions;
    private HashTable<string,Icon> file_icon_cache;

    construct
    {
      var sections_model = new Dee.SharedModel(
                                 "com.canonical.Unity.ApplicationsPlace.SectionsModel",
                                 2, typeof (string), typeof (string));

      var groups_model = new Dee.SharedModel(
                                 "com.canonical.Unity.ApplicationsPlace.GroupsModel",
                                 3, typeof (string), typeof (string),
                                 typeof (string));

      var global_groups_model = new Dee.SharedModel(
                                 "com.canonical.Unity.ApplicationsPlace.GlobalGroupsModel",
                                 3, typeof (string), typeof (string),
                                 typeof (string));

      var results_model = new Dee.SharedModel(
                                 "com.canonical.Unity.ApplicationsPlace.ResultsModel",
                                 6, typeof (string), typeof (string),
                                 typeof (uint), typeof (string),
                                 typeof (string), typeof (string));

      var global_results_model = new Dee.SharedModel(
                                 "com.canonical.Unity.ApplicationsPlace.GlobalResultsModel",
                                 6, typeof (string), typeof (string),
                                 typeof (uint), typeof (string),
                                 typeof (string), typeof (string));

      section_queries = new Gee.ArrayList<string> ();
      prepare_section_queries();

      applications = new EntryInfo ("/com/canonical/unity/applicationsplace/applications");
      applications.sections_model = sections_model;
      applications.entry_renderer_info.groups_model = groups_model;
      applications.entry_renderer_info.results_model = results_model;
      applications.global_renderer_info.groups_model = global_groups_model;
      applications.global_renderer_info.results_model = global_results_model;

      populate_sections ();
      populate_groups ();

      applications.icon = @"$(Config.PREFIX)/share/unity/applications.png";

      log = new Zeitgeist.Log();
      index = new Zeitgeist.Index();
      pkgsearcher = new Unity.Package.Searcher ();
      
      /* Image file extensions in order of popularity */
      image_extensions = new Gee.ArrayList<string> ();
      image_extensions.add ("png");
      image_extensions.add ("xpm");
      image_extensions.add ("svg");
      image_extensions.add ("tiff");
      image_extensions.add ("ico");
      image_extensions.add ("tif");
      image_extensions.add ("jpg");
      
      file_icon_cache = new HashTable<string,Icon>(str_hash, str_equal);
      
      /* Listen for section changes */
      applications.notify["active-section"].connect (
        (obj, pspec) => {
          applications.entry_renderer_info.results_model.clear();
          update_entry_results_model.begin();
        }
      );

      /* Listen for changes to the place entry search */
      applications.notify["active-search"].connect (
        (obj, pspec) => {
          var _results_model = applications.entry_renderer_info.results_model;
          var _groups_model = applications.entry_renderer_info.groups_model;
          search_async.begin (applications.active_search,
                              (Section)applications.active_section,
                              _results_model, _groups_model);
        }
      );

      /* Listen for changes to the global search */
      applications.notify["active-global-search"].connect (
        (obj, pspec) => {
          var _results_model = applications.global_renderer_info.results_model;
          var _groups_model = applications.global_renderer_info.groups_model;
          search_async.begin (applications.active_global_search,
                              Section.ALL_APPLICATIONS,
                              _results_model, _groups_model);
        }
      );

      /* We should not do anything with the results model
       * until we receieve the 'ready' signal */
      results_model.ready.connect (
        (model) => { update_entry_results_model.begin(); }
      );

      sections_model.connect ();
      groups_model.connect ();
      global_groups_model.connect ();
      results_model.connect ();
      global_results_model.connect ();

      /* The last thing we do is export the controller. Once that is up,
       * clients will expect the SharedModels to work */
      control = new Unity.Place.Controller ("/com/canonical/unity/applicationsplace");
      control.add_entry (applications);
      try {
        control.export ();
      } catch (DBus.Error error) {
        critical ("Failed to export DBus service for '%s': %s",
                  control.dbus_path, error.message);
      }

    }

    private void populate_sections ()
    {
      var sections = applications.sections_model;

      if (sections.get_n_rows() != 0)
        {
          critical ("The sections model should be empty before initial population");
          sections.clear ();
        }

      sections.append (SectionsColumn.DISPLAY_NAME, "All Applications",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Accessories",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Games",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Internet",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Media",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "Office",
                       SectionsColumn.ICON_HINT, "", -1);
      sections.append (SectionsColumn.DISPLAY_NAME, "System",
                       SectionsColumn.ICON_HINT, "", -1);
    }

    private void populate_groups ()
    {
      var groups = applications.entry_renderer_info.groups_model;

      if (groups.get_n_rows() != 0)
        {
          critical ("The groups model should be empty before initial population");
          groups.clear ();
        }

      // FIXME: Bigger tiles for most used group
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "Most Used",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "Installed",
                     GroupsColumn.ICON_HINT, "", -1);
      groups.append (GroupsColumn.RENDERER, "UnityDefaultRenderer",
                     GroupsColumn.DISPLAY_NAME, "Available",
                     GroupsColumn.ICON_HINT, "", -1);
    }

    private void prepare_section_queries ()
    {
      section_queries.add (""); //ALL_APPLICATIONS
      section_queries.add ("category:Utility"); //ACCESSORIES
      section_queries.add ("category:Game"); //GAMES
      section_queries.add ("category:Network"); //INTERNET
      section_queries.add ("(category:AudioVideo OR category:Graphics)"); //MEDIA
      section_queries.add ("category:Office"); //OFFICE
      section_queries.add ("(category:System OR category:Settings)"); //SYSTEM
    }

    private string prepare_search_string (Search search)
    {
      var s = search.get_search_string ();

      if (s.has_suffix (" "))
        s = s.strip ();

      if (!s.has_suffix ("*"))
        s = s + "*";

      return @"app:($s)";
    }

    private async void search_async (Search search,
                                     Section section,
                                     Dee.Model results_model,
                                     Dee.Model groups_model)
    {
      if (search == null ||
          search.get_search_string () == null ||
          search.get_search_string () == "")
        {
          update_entry_results_model.begin ();
          return;
        }      
      
      // FIXME: Restrict to active section
      //var templates = section_templates.get((int)section);
      var templates = new PtrArray ();
      var search_string = prepare_search_string (search);

      try {
        var results = yield index.search (search_string,
                                          new Zeitgeist.TimeRange.anytime(),
                                          templates,
                                          Zeitgeist.StorageState.ANY,
                                          20,
                                          Zeitgeist.ResultType.MOST_POPULAR_ACTOR,
                                          null);

        results_model.clear ();
        Unity.ApplicationsPlace.append_events_with_group (results,
                                                          results_model,
                                                          Group.MOST_USED);

        debug ("Found %u/%u matches for search '%s'",
               results.size (), results.estimated_matches (),
               search.get_search_string ());

      } catch (GLib.Error e) {
        warning ("Error performing global search '%s': %s",
                 search.get_search_string (), e.message);
      }
      
      // FIXME: We could parallelize the queries to the packages and zg
      update_pkg_search (search, section, results_model);
    }

    private async void update_entry_results_model ()
    {
      var search = applications.active_search;
      var section = applications.active_section;
      var sections_model = applications.sections_model;
      var results_model = applications.entry_renderer_info.results_model;
      var groups_model = applications.entry_renderer_info.groups_model;

      if (Section.LAST_SECTION != sections_model.get_n_rows())
        {
          critical ("Section model malformed");
          return;
        }

      if (section > Section.LAST_SECTION)
        {
          critical ("Active section out of bounds: %u", section);
          return;
        }

      /* Create a template that matches anything */
      var templates = new PtrArray.sized(1);

      /* Find most popular apps */
      Zeitgeist.ResultSet? popular_apps = null;
      try {
        popular_apps = yield log.find_events (
                                        new Zeitgeist.TimeRange.anytime(),
                                        templates,
                                        Zeitgeist.StorageState.ANY,
                                        6,
                                        Zeitgeist.ResultType.MOST_POPULAR_ACTOR,
                                        null);
      } catch (GLib.Error e) {
        warning ("Error finding popular applications: %s", e.message);
      }

      /* Clear the model only after we have the search results to minize
       * the time we show an empty model */
      results_model.clear ();

      if (popular_apps != null)
        {
          ApplicationsPlace.append_events_with_group (popular_apps,
                                                      results_model,
                                                      Group.MOST_USED);
        }

      /* List installed app by section */
      update_pkg_search (search, (Section) section, results_model);
    }
    
    private string prepare_pkg_search_string (Search? search, Section section)
    {       
      if (search == null)
        {
          if (section == Section.ALL_APPLICATIONS)
            return "type:Application";
          else
            return @"type:Application AND $(section_queries.get(section))";
        }
      else
        {
          var s = search.get_search_string ();

          s = s.strip ();

          if (!s.has_suffix ("*"))
            s = s + "*";

          if (section == Section.ALL_APPLICATIONS)
            return @"($s) AND type:Application";
          else
            return @"($s) AND type:Application AND $(section_queries.get(section))";
        }
    }
    
    private Icon find_pkg_icon (Unity.Package.PackageInfo pkginfo)
    {       
      /* If the app is already installed we should be able to pull the
       * icon from the theme*/
      if (pkginfo.installed)
        return new ThemedIcon (pkginfo.icon);
      
      /* App is not installed - we need to find the right icon in the bowels
       * of the software center */
      /* FIXME: Cache the image locations because this requires a lot of stat()ing */
      if (pkginfo.icon.has_prefix ("/"))
        {
          return new FileIcon (File.new_for_path (pkginfo.icon));
        }
      else
        {
          Icon icon = file_icon_cache.lookup (pkginfo.icon);
          
          if (icon != null)
            return icon;
        
          foreach (var ext in image_extensions)
          {
            string path = @"/usr/share/app-install/icons/$(pkginfo.icon).$(ext)";
            File f = File.new_for_path (path);
            if (f.query_exists (null))
              {
                /* Got it! Cache the icon path and return the icon */
                icon = new FileIcon (f);
                file_icon_cache.insert (pkginfo.icon, icon);
                return icon;
              }
          }
        }
      
      /* Cache the fact that we couldn't find this icon */
      var icon = new ThemedIcon ("applications-other");
      file_icon_cache.insert (pkginfo.icon, icon);
      
      return icon;
    }

    private bool update_pkg_search (Search? search,
                                    Section section,
                                    Dee.Model model)
    {
      string search_string = prepare_pkg_search_string (search, section);
      
      Timer timer = new Timer ();
      var results = pkgsearcher.search (search_string);
      foreach (var pkginfo in results.results)
      {
        if (pkginfo.desktop_file == null)
          continue;
        
        AppInfo app = AppInfoManager.get_instance().lookup (pkginfo.desktop_file);
        if (!(app is AppInfo))
          {
            warning ("Unable to read: %s", pkginfo.desktop_file);
            continue;
          }
        
        string uri;
        if (pkginfo.installed)
          {
            var f = File.new_for_path (pkginfo.desktop_file);
            uri = @"application://$(f.get_basename())";
          }
        else
          uri = @"file://$(pkginfo.desktop_file)";
          
        Icon icon = find_pkg_icon (pkginfo);
        
        print (@" * $(pkginfo.application_name) - $(icon.to_string())\n");
        model.append (ResultsColumn.URI, uri,
                      ResultsColumn.ICON_HINT, icon.to_string (),
                      ResultsColumn.GROUP_ID, pkginfo.installed ? Group.INSTALLED : Group.AVAILABLE,
                      ResultsColumn.MIMETYPE, "application/x-desktop",
                      ResultsColumn.DISPLAY_NAME, app.get_display_name (),
                      ResultsColumn.COMMENT, app.get_description (),
                      -1);
      }
      
      timer.stop ();
      debug ("Found %i apps in %fms for query: %s",
             results.num_hits, timer.elapsed ()/1000, search_string);
      
      return false;
    }

  } /* END: class Daemon */

  /* Appends the actors from a set of Zeitgeist.Events to our Dee.Model
   * assuming that these events are already sorted */
  public void append_events_with_group (ResultSet events,
                                        Dee.Model results,
                                        uint group_id)
  {
    foreach (var ev in events)
      {
        AppInfo app = Utils.get_app_info_for_actor (ev.get_actor ());
        string app_uri = @"application://$(app.get_id())";
        results.append (ResultsColumn.URI, app_uri,
                        ResultsColumn.ICON_HINT, app.get_icon().to_string(),
                        ResultsColumn.GROUP_ID, group_id,
                        ResultsColumn.MIMETYPE, "application/x-desktop",
                        ResultsColumn.DISPLAY_NAME, app.get_display_name (),
                        ResultsColumn.COMMENT, app.get_description (),
                        -1);
      }
  }

} /* namespace */
