/*
 * Copyright (C) 2010 Canonical Ltd
   *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Mikkel Kamstrup Erlandsen <mikkel.kamstrup@canonical.com>
 *
 */

/*
 * This file contains C bindings written in C++ to hook up with the
 * Xapian indexes generated by Ubuntu Software Center and apt-xapian.
 * It also hooks up to Apt's cache via libapt-pkg to check whether or not
 * packages are installed.
 *
 * In the future we may be able to drop the depency on libapt-pkg if the Xapian
 * indexes include some installation metadata.
 */

using namespace std;

#include <apt-pkg/pkgcachegen.h>
#include <apt-pkg/cachefile.h>
#include <apt-pkg/cacheiterators.h>
#include <apt-pkg/progress.h>
#include <apt-pkg/init.h>

#include <xapian.h>
#include <iostream>

#define SOFTWARE_CENTER_INDEX "/var/cache/software-center/xapian"
#define AXI_INDEX "/var/lib/apt-xapian-index/index"
#define QUERY_PARSER_FLAGS Xapian::QueryParser::FLAG_BOOLEAN|Xapian::QueryParser::FLAG_PHRASE|Xapian::QueryParser::FLAG_LOVEHATE|Xapian::QueryParser::FLAG_WILDCARD

// values used in the database
#define XAPIAN_VALUE_APPNAME 170
#define XAPIAN_VALUE_PKGNAME 171
#define XAPIAN_VALUE_ICON 172
#define XAPIAN_VALUE_GETTEXT_DOMAIN 173
#define XAPIAN_VALUE_ARCHIVE_SECTION 174
#define XAPIAN_VALUE_ARCHIVE_ARCH 175
#define XAPIAN_VALUE_POPCON 176
#define XAPIAN_VALUE_SUMMARY 177
#define XAPIAN_VALUE_ARCHIVE_CHANNEL 178
#define XAPIAN_VALUE_DESKTOP_FILE 179

#include "unity-package-search.h"

struct _UnityPackageSearcher
{
  pkgCache *cache;
  Xapian::Database *db;
  Xapian::Enquire *enquire;
  Xapian::QueryParser *query_parser;
};

UnityPackageSearcher*
unity_package_searcher_new ()
{
  UnityPackageSearcher *searcher;

  searcher = g_new0 (UnityPackageSearcher, 1);

  // Xapian initialization
  try {
    // Load the Xapian database
    Xapian::Database *sc_db = new Xapian::Database (SOFTWARE_CENTER_INDEX);
    Xapian::Database *axi_db = new Xapian::Database (AXI_INDEX);
    sc_db->add_database (*axi_db);
    searcher->db = sc_db;
    delete axi_db;

    // Start an enquire session
    Xapian::Enquire *enquire = new Xapian::Enquire (*sc_db);
    searcher->enquire = enquire;

    // Create query parser
    Xapian::QueryParser *query_parser = new Xapian::QueryParser ();
    query_parser->add_prefix ("section", "AE");
    query_parser->add_prefix ("type", "AT");
    query_parser->add_prefix ("category", "AC");
    query_parser->add_prefix ("name", "AA");
    query_parser->add_prefix ("pkgname", "AP");
    query_parser->set_default_op (Xapian::Query::OP_AND);
    query_parser->set_database (*sc_db);
    searcher->query_parser = query_parser;
  } catch(const Xapian::Error &error) {
    cerr << "Error loading package indexes: "  << error.get_msg() << endl;
    return NULL;
  }

  // AptPkg initialization
  // FIXME: We should avoid calling the pkgInit* functions more than once
  pkgInitConfig(*_config);
  pkgInitSystem(*_config, _system);

  pkgSourceList list;
  list.ReadMainList();
  OpTextProgress progress;
  MMap *map;
  bool res = pkgMakeStatusCache(list, progress, &map, true);
  progress.Done ();
  pkgCache *cache = new pkgCache (map);
  searcher->cache = cache;

  return searcher;
}

void
unity_package_searcher_free (UnityPackageSearcher *searcher)
{
  g_return_if_fail (searcher != NULL);

  delete searcher->db;
  delete searcher->enquire;
  delete searcher->query_parser;
  delete searcher->cache;
  g_free (searcher);
}

UnityPackageSearchResult*
unity_package_searcher_search (UnityPackageSearcher *searcher,
                               const gchar          *search_string)
{
  UnityPackageSearchResult* result;

  g_return_val_if_fail (searcher != NULL, NULL);
  g_return_val_if_fail (search_string != NULL, NULL);

  string _search_string (search_string);
  Xapian::Query query = searcher->query_parser->parse_query (_search_string, QUERY_PARSER_FLAGS);
  cout << "Performing query `" << query.get_description() << "'" << endl;

  // Perform search
  searcher->enquire->set_query(query);
  Xapian::MSet matches =
    searcher->enquire->get_mset(0, searcher->db->get_doccount ());	

  // Retrieve the results, note that we build the result->results
  // list in reverse order and then reverse it before we return it
  result = g_slice_new0 (UnityPackageSearchResult);
  result->num_hits = matches.get_matches_estimated ();
  for (Xapian::MSetIterator i = matches.begin();
       i != matches.end();
       ++i)
    {
      UnityPackageInfo *pkginfo = g_slice_new0 (UnityPackageInfo);
      Xapian::Document doc = i.get_document();

      string pkgname = doc.get_value (XAPIAN_VALUE_PKGNAME);
      pkginfo->package_name = g_strdup (pkgname.c_str ());
      if (pkgname != "")
        {
          pkgCache::PkgIterator pkg = searcher->cache->FindPkg(pkgname);

          if (!pkg.end() && pkg->CurrentState == pkgCache::State::Installed)
            pkginfo->installed = TRUE;
          else
            pkginfo->installed = FALSE;
        }
      else
        pkginfo->installed = FALSE;

      string appname = doc.get_value (XAPIAN_VALUE_APPNAME);
      pkginfo->application_name = g_strdup (appname.c_str ());

      string desktop_file = doc.get_value (XAPIAN_VALUE_DESKTOP_FILE);
      pkginfo->desktop_file = g_strdup (desktop_file.c_str ());

      string icon = doc.get_value (XAPIAN_VALUE_ICON);
      pkginfo->icon = g_strdup (icon.c_str ());

      result->results = g_slist_prepend (result->results, pkginfo);
    }

  result->results = g_slist_reverse (result->results);
  return result;
}

// extern "C"
static void _free_package_info (gpointer pkg, gpointer data)
{
  UnityPackageInfo *pkginfo = (UnityPackageInfo*) pkg;

  g_free (pkginfo->package_name);
  g_free (pkginfo->application_name);
  g_free (pkginfo->desktop_file);
  g_free (pkginfo->icon);

  g_slice_free (UnityPackageInfo, pkg);
}

void
unity_package_search_result_free (UnityPackageSearchResult *result)
{
  g_return_if_fail (result != NULL);

  g_slist_foreach (result->results, _free_package_info, NULL);
  g_slist_free (result->results);
  g_slice_free (UnityPackageSearchResult, result);
}

